# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['multisort']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'multisort',
    'version': '0.1.4',
    'description': 'NoneType Safe Multi Column Sorting For Python',
    'long_description': '## `multisort` - NoneType Safe Multi Column Sorting For Python\n\nSimplified multi-column sorting of lists of tuples, dicts, lists or objects that are NoneType safe.\n\n### Installation\n\n```\npython3 -m pip install multisort\n```\n\n### Dependencies\nNone\n\n### Performance\nAverage over 10 iterations with 1000 rows.\nTest | Secs\n---|---\nsuperfast|0.0005\nmultisort|0.0035\npandas|0.0079\ncmp_func|0.0138\nreversor|0.037\n\nHands down the fastest is the `superfast` methdology shown below. You do not need this library to accomplish this as its just core python.\n\n`multisort` from this library gives reasonable performance for large data sets; eg. its better than pandas up to about 5,500 records. It is also much simpler to read and write, and it has error handling that does its best to give useful error messages.\n\n### Note on `NoneType` and sorting\nIf your data may contain None, it would be wise to ensure your sort algorithm is tuned to handle them. This is because sorted uses `<` comparisons; which is not supported by `NoneType`. For example, the following error will result: `TypeError: \'>\' not supported between instances of \'NoneType\' and \'str\'`. All examples given on this page are tuned to handle `None` values.\n\n### Methodologies\nMethod|Descr|Notes\n---|---|---\nmultisort|Simple one-liner designed after `multisort` [example from python docs](https://docs.python.org/3/howto/sorting.html#sort-stability-and-complex-sorts)|Second fastest of the bunch but most configurable and easy to read.\ncmp_func|Multi column sorting in the model `java.util.Comparator`|Reasonable speed|Enable multi column sorting with column specific reverse sorting|Medium speed. [Source](https://stackoverflow.com/a/56842689/286807)\nsuperfast|NoneType safe sample implementation of multi column sorting as mentioned in [example from python docs](https://docs.python.org/3/howto/sorting.html#sort-stability-and-complex-sorts)|Fastest by orders of magnitude but a bit more complex to write.\n\n\n<br>\n\n### Dictionary Examples\nFor data:\n```\nrows_before = [\n     {\'idx\': 0, \'name\': \'joh\', \'grade\': \'C\', \'attend\': 100}\n    ,{\'idx\': 1, \'name\': \'jan\', \'grade\': \'a\', \'attend\': 80}\n    ,{\'idx\': 2, \'name\': \'dav\', \'grade\': \'B\', \'attend\': 85}\n    ,{\'idx\': 3, \'name\': \'bob\' , \'grade\': \'C\', \'attend\': 85}\n    ,{\'idx\': 4, \'name\': \'jim\' , \'grade\': \'F\', \'attend\': 55}\n    ,{\'idx\': 5, \'name\': \'joe\' , \'grade\': None, \'attend\': 55}\n]\n```\n\n### `multisort`\nSort rows_before by _grade_, descending, then _attend_, ascending and put None first in results:\n```\nfrom multisort import multisort, mscol\nrows_sorted = multisort(rows_before, [\n        mscol(\'grade\', reverse=False),\n        \'attend\'\n])\n```\n\n-or- without `mscol`\n\n```\nfrom multisort import multisort\nrows_sorted = multisort(rows_before, [\n        (\'grade\', False),\n        \'attend\'\n])\n```\n\nSort rows_before by _grade_, descending, then _attend_ and call upper() for _grade_:\n```\nfrom multisort import multisort, mscol\nrows_sorted = multisort(rows_before, [\n        mscol(\'grade\', reverse=False, clean=lambda s: None if s is None else s.upper()),\n        \'attend\'\n])\n```\n\n-or- without `mscol`\n```\nfrom multisort import multisort\nrows_sorted = multisort(rows_before, [\n        (\'grade\', False, lambda s: None if s is None else s.upper()),\n        \'attend\'\n])\n```\n\n\n`multisort` parameters:\noption|dtype|description\n---|---|---\n`rows`|int or str|Key to access data. int for tuple or list\n`spec`|str, int, list|Sort specification. Can be as simple as a column key / index or `mscol`\n`reverse`|bool|Reverse order of final sort (defalt = False)\n\n\n`spec` entry options:\noption|position|dtype|description\n---|---|---|---\n`key`|0|int or str|Key to access data. int for tuple or list\n`reverse`|1|bool|Reverse sort of column\n`clean`|2|func|Function / lambda to clean the value. These calls can cause a significant slowdown.\n`default`|3|any|Value to substitute if required==False and key does not exist or None is found. Can be used to achive similar functionality to pandas `na_position`\n`required`|4|bool|Default True. If False, will substitute None or default if key not found (not applicable for list or tuple rows)\n\n\n\\* `spec` entries can be passed as:\n&nbsp;&nbsp;&nbsp;&nbsp;type|description\n---|---\n&nbsp;&nbsp;&nbsp;&nbsp;`String`|Column name\n&nbsp;&nbsp;&nbsp;&nbsp;`tuple`|Tuple of 1 or more `spec` options in their order as listed (see `position`)\n&nbsp;&nbsp;&nbsp;&nbsp;`mscol()`|Importable helper to aid in readability. Suggested for three or more of the options.\n\n\n<br><br>\n\n\n### `sorted` with `cmp_func`\nSort rows_before by _grade_, descending, then _attend_ and call upper() for _grade_:\n```\ndef cmp_student(a,b):\n    k=\'grade\'; va=a[k]; vb=b[k]\n    if va != vb: \n        if va is None: return -1\n        if vb is None: return 1\n        return -1 if va > vb else 1\n    k=\'attend\'; va=a[k]; vb=b[k]; \n    if va != vb: return -1 if va < vb else 1\n    return 0\nrows_sorted = sorted(rows_before, key=cmp_func(cmp_student), reverse=True)\n```\n\n<br>\n\n### For reference: `superfast` methodology with list of dicts:\n```\ndef key_grade(student):\n    grade = student[\'grade\']\n    return grade is None, grade\ndef key_attend(student):\n    attend = student[\'attend\']\n    return attend is None, attend\nstudents_sorted = sorted(students, key=key_attend)\nstudents_sorted.sort(key=key_grade, reverse=True)\n```\n\n<br>\n\n### Object Examples\nFor data:\n```\nclass Student():\n    def __init__(self, idx, name, grade, attend):\n        self.idx = idx\n        self.name = name\n        self.grade = grade\n        self.attend = attend\n    def __str__(self): return f"name: {self.name}, grade: {self.grade}, attend: {self.attend}"\n    def __repr__(self): return self.__str__()\n\nrows_before = [\n     Student(0, \'joh\', \'C\', 100)\n    ,Student(1, \'jan\', \'a\', 80)\n    ,Student(2, \'dav\', \'B\', 85)\n    ,Student(3, \'bob\', \'C\', 85)\n    ,Student(4, \'jim\', \'F\', 55)\n    ,Student(5, \'joe\', None, 55)\n]\n```\n<br>\n\n### `multisort`\n(Same syntax as with Dictionary example above)\n\n<br>\n\n### `sorted` with `cmp_func`\nSort rows_before by _grade_, descending, then _attend_ and call upper() for _grade_:\n```\ndef cmp_student(a,b):\n    if a.grade != b.grade: \n        if a.grade is None: return -1\n        if b.grade is None: return 1\n        return -1 if a.grade > b.grade else 1\n    if a.attend != b.attend: \n        return -1 if a.attend < b.attend else 1\n    return 0\nrows_sorted = sorted(rows_before, key=cmp_func(cmp_student), reverse=True)\n```\n\n\n### List / Tuple Examples\nFor data:\n```\nrows_before = [\n     (0, \'joh\', \'a\'  , 100)\n    ,(1, \'joe\', \'B\'  , 80)\n    ,(2, \'dav\', \'A\'  , 85)\n    ,(3, \'bob\', \'C\'  , 85)\n    ,(4, \'jim\', None , 55)\n    ,(5, \'jan\', \'B\'  , 70)\n]\n(COL_IDX, COL_NAME, COL_GRADE, COL_ATTEND) = range(0,4)\n```\n\n<br>\n\n### `multisort`\n(Same syntax as with Dictionary example above)\n\n<br>\n\n### `sorted` with `cmp_func`\nSort rows_before by _grade_, descending, then _attend_ and call upper() for _grade_:\n```\ndef cmp_student(a,b):\n    k=COL_GRADE; va=a[k]; vb=b[k]\n    if va != vb: \n        if va is None: return -1\n        if vb is None: return 1\n        return -1 if va > vb else 1\n    k=COL_ATTEND; va=a[k]; vb=b[k]; \n    if va != vb: \n        return -1 if va < vb else 1\n    return 0\nrows_sorted = sorted(rows_before, key=cmp_func(cmp_student), reverse=True)\n```\n\n<br><br>\n\n\n### `multisort` library Test / Sample files (/tests)\nName|Descr|Other\n---|---|---\ntests/test_multisort.py|multisort unit tests|- \ntests/performance_tests.py|Tunable performance tests using asyncio | requires pandas\ntests/hand_test.py|Hand testing|-\n',
    'author': 'Timothy C. Quinn',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/multisort',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7.9,<4.0.0',
}


setup(**setup_kwargs)
